/*
 This project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Deviation is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Deviation.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Uncomment define below to enable packet loss telemetry. Also add
   H1 and SymaX protocols to TELEMETRY_SetTypeByProtocol to
   set type to DSM.
   */
//#define H1_TELEMETRY


#ifdef MODULAR
  //Allows the linker to properly relocate
  #define H1_Cmds PROTO_Cmds
  #pragma long_calls
#endif
#include "common.h"
#include "interface.h"
#include "mixer.h"
#include "config/model.h"
#include "config/tx.h" // for Transmitter
#include "music.h"
#include "telemetry.h"

#ifdef MODULAR
  //Some versions of gcc applythis to definitions, others to calls
  //So just use long_calls everywhere
  //#pragma long_calls_off
  extern unsigned _data_loadaddr;
  const unsigned long protocol_type = (unsigned long)&_data_loadaddr;
#endif

#ifdef PROTO_HAS_NRF24L01

#include "iface_nrf24l01.h"

#ifdef EMULATOR
#define USE_FIXED_MFGID
#define BIND_COUNT 10
#define PACKET_PERIOD   1000     // Adjust timeouts for reasonable emulator printouts
#define INITIAL_WAIT     500
#define dbgprintf printf
#else


// these are the real ones!
#define BIND_COUNT 32 // how many times to check the status register for an ack
#define PACKET_PERIOD   5000     // Timeout for callback in uSec, 5ms=5000us for H1
#define INITIAL_WAIT   50000     // Initial wait before starting callbacks

//printf inside an interrupt handler is really dangerous
//this shouldn't be enabled even in debug builds without explicitly
//turning it on
#define dbgprintf if(0) printf
#endif

#define PACKET_CHKTIME   250     // Time to wait if packet not yet acknowledged or timed out    

#define PACKET_SIZE 0x1d

// these don't matter yet
#define FLAG_FLIP   0x0F
#define FLAG_LIGHT  0x10

// For code readability
enum {
    CHANNEL1 = 0,
    CHANNEL2,
    CHANNEL3,
    CHANNEL4,
    CHANNEL5,
    CHANNEL6,
    CHANNEL7,
    CHANNEL8,
    CHANNEL9,
    CHANNEL10
};


static u8 packet[PACKET_SIZE];
static u16 counter;
static u32 packet_counter;
static u8 tx_power;
static u8 throttle, rudder, elevator, aileron, flags;
static u8 rudder_trim, elevator_trim, aileron_trim;
//static u8 rx_tx_addr[5];

static u8 data_channel;

uint16_t crc_kermit =0;

static u8 phase;
enum {
    H1_INIT1 = 0,
    H1_BIND2,
    H1_BIND3,
    H1_DATA
};


static const char * const h1_opts[] = {
  _tr_noop("Format"),  _tr_noop("H1"), _tr_noop("E022"), NULL,
  _tr_noop("Bind Frequency"),  _tr_noop("2412 MHz"), _tr_noop("2410 MHz"), _tr_noop("2408 MHz"), _tr_noop("2421 MHz"), _tr_noop("2408 MHz"), _tr_noop("2419 MHz"), _tr_noop("2411 MHz"), NULL,
#ifdef H1_TELEMETRY
  _tr_noop("Telemetry"),  _tr_noop("Off"), _tr_noop("On"), NULL,
#endif
  NULL
};
enum {
    PROTOOPTS_FORMAT = 0,
    PROTOOPTS_PAYLOAD,
#ifdef H1_TELEMETRY
    PROTOOPTS_TELEMETRY,
#endif
    LAST_PROTO_OPT,
};
ctassert(LAST_PROTO_OPT <= NUM_PROTO_OPTS, too_many_protocol_opts);

#define FORMAT_H1   0
#define FORMAT_SKYWLKR 1
#define FORMAT_XINXUN  2

#ifdef H1_TELEMETRY
#define TELEM_OFF 0
#define TELEM_ON 1
#endif

// Bit vector from bit position
#define BV(bit) (1 << bit)

#define TRUE 1
#define FALSE 0

/*
a9 dd 7d 09: bind (ch 0a)
13 17 26 3c: channels

7a bc ff 37: bind (ch 08)
11 15 30 3D: channels
1a e1 84 37 00 01 02 88 00 00 7f 7f 32 20 40 40 00 d0: data

78 be df 57: bind (ch 06)
0f 17 2e 3f: channels
18 e3 64 57 02 01 02 88 00 00 7f 7f 32 20 40 40 00 d0

c1 75 aa 8c: bind (ch 13)
0d 19 26 38: channels
61 9a 2f 8c 01 01 02 88 00 00 7f 7f 32 20 40 40 00 d0

2d 09 3e f8: bind (ch 06)
10 16 24 3b: channels
cd 2e c3 f8 02 01 02 88 00 00 7f 7f 32 20 40 40 00 d0

65 d1 2c 0a: bind (ch 11)
0b 1b 30 3d: channels
05 f6 b1 0a 00 01 02 88 00 00 7f 7f 32 20 40 40 00 d0

a8 8e ed 49: bind (ch 09)
12 23 2d 40: channels
48 b3 72 49 00 04 02 88 00 00 7e 81 32 20 40 40 00 d1
*/

const u8 bind_channels[] = {
  0x0a,
  0x08,
  0x06,
  0x13,
  0x06,
  0x11,
  0x09,
};

const u8 bind_payloads[7][4] = {
	{ 0xa9, 0xdd, 0x7d, 0x09 },
	{ 0x7a, 0xbc, 0xff, 0x37 },
	{ 0x78, 0xbe, 0xdf, 0x57 },
	{ 0xc1, 0x75, 0xaa, 0x8c },
	{ 0x2d, 0x09, 0x3e, 0xf8 },
	{ 0x65, 0xd1, 0x2c, 0x0a },
	{ 0xa8, 0x8e, 0xed, 0x49 },
};

const u8 data_channels[7][4] = {
	{ 0x13, 0x17, 0x26, 0x3c },
	{ 0x11, 0x15, 0x30, 0x3D },
	{ 0x0f, 0x17, 0x2e, 0x3f },
	{ 0x0d, 0x19, 0x26, 0x38 },
	{ 0x10, 0x16, 0x24, 0x3b },
	{ 0x0b, 0x1b, 0x30, 0x3d },
	{ 0x12, 0x23, 0x2d, 0x40 },
};

const u8 data_payloads[7][4] = {
  { 0x49, 0x02, 0x02, 0x09 },
	{ 0x1a, 0xe1, 0x84, 0x37 },
	{ 0x18, 0xe3, 0x64, 0x57 },
	{ 0x61, 0x9a, 0x2f, 0x8c },
	{ 0xcd, 0x2e, 0xc3, 0xf8 },
	{ 0x05, 0xf6, 0xb1, 0x0a },
	{ 0x48, 0xb3, 0x72, 0x49 },
};

uint8_t reverse(unsigned char b) {
   b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
   b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
   b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
   return b;
}

#define                 P_KERMIT    0x8408
static int              crc_tabkermit_init      = FALSE;
static unsigned short   crc_tabkermit[256];

    /*******************************************************************\
    *                                                                   *
    *   static void init_crckermit_tab( void );                         *
    *                                                                   *
    *   The function init_crckermit_tab() is used to fill the array     *
    *   for calculation of the CRC Kermit with values.                  *
    *                                                                   *
    \*******************************************************************/

static void init_crckermit_tab( void ) {

    int i, j;
    unsigned short crc, c;

    for (i=0; i<256; i++) {

        crc = 0;
        c   = (unsigned short) i;

        for (j=0; j<8; j++) {

            if ( (crc ^ c) & 0x0001 ) crc = ( crc >> 1 ) ^ P_KERMIT;
            else                      crc =   crc >> 1;

            c = c >> 1;
        }

        crc_tabkermit[i] = crc;
    }

    crc_tabkermit_init = TRUE;

}  /* init_crckermit_tab */


    /*******************************************************************\
    *                                                                   *
    *   unsigned short update_crc_kermit( unsigned short crc, char c ); *
    *                                                                   *
    *   The function update_crc_kermit calculates a  new  CRC value     *
    *   based  on  the  previous value of the CRC and the next byte     *
    *   of the data to be checked.                                      *
    *                                                                   *
    \*******************************************************************/

unsigned short update_crc_kermit( unsigned short crc, char c ) {

    unsigned short tmp, short_c;

    short_c = 0x00ff & (unsigned short) c;

    if ( ! crc_tabkermit_init ) init_crckermit_tab();

    tmp =  crc       ^ short_c;
    crc = (crc >> 8) ^ crc_tabkermit[ tmp & 0xff ];

    return crc;

}  /* update_crc_kermit */




static u8 convert_channel(u8 num)
{
    // Sometimes due to imperfect calibration or mixer settings
    // throttle can be less than CHAN_MIN_VALUE or larger than
    // CHAN_MAX_VALUE. As we have no space here, we hard-limit
    // channels values by min..max range
    s32 ch = Channels[num];
    if (ch < CHAN_MIN_VALUE) {
        ch = CHAN_MIN_VALUE;
    } else if (ch > CHAN_MAX_VALUE) {
        ch = CHAN_MAX_VALUE;
    }

    return (u8) (((ch * 0xFF / CHAN_MAX_VALUE) + 0x100) >> 1);
}


static void read_controls(u8* throttle, u8* rudder, u8* elevator, u8* aileron,
                          u8* flags, u8* rudder_trim, u8* elevator_trim, u8* aileron_trim)
{
    // Protocol is registered AETRF, that is
    // Aileron is channel 1, Elevator - 2, Throttle - 3, Rudder - 4, Flip control - 5

    // Channel 3
    *throttle = convert_channel(CHANNEL3);

    // Channel 4
    *rudder = convert_channel(CHANNEL4);
    *rudder_trim = (0xff - *rudder) >> 1;

    // Channel 2
    *elevator = convert_channel(CHANNEL2);
    *elevator_trim = *elevator >> 1;

    // Channel 1
    *aileron = 0xff - convert_channel(CHANNEL1);
    *aileron_trim = *aileron >> 1;

    // Channel 5
    if (Channels[CHANNEL5] <= 0)
      *flags &= ~FLAG_FLIP;
    else
      *flags |= FLAG_FLIP;

    // Channel 6
    if (Channels[CHANNEL6] <= 0)
      *flags &= ~FLAG_LIGHT;
    else
      *flags |= FLAG_LIGHT;


    dbgprintf("ail %3d+%3d, ele %3d+%3d, thr %3d, rud %3d+%3d, flip enable %d\n",
            *aileron, *aileron_trim, *elevator, *elevator_trim, *throttle,
            *rudder, *rudder_trim, *flags);
}





// Packet ack status values
enum {
    PKT_PENDING = 0,
    PKT_ACKED,
};




static u8 packet_ack()
{
    dbgprintf("Ack counter: %02d\n", counter);
    if (NRF24L01_ReadReg(NRF24L01_07_STATUS) & BV(NRF24L01_07_RX_DR)) {
    NRF24L01_WriteReg(NRF24L01_07_STATUS, BV(NRF24L01_07_RX_DR));
#ifdef EMULATOR // there is no nRF in my pc
        dbgprintf("ACK pending (pretend)\n");
        return PKT_PENDING;
#else
        return PKT_ACKED;
#endif
    }
    return PKT_PENDING;
}


static void send_packet(u8 bind)
{
    if (bind) {

        dbgprintf("Sending bind packet");
        for (int i=0; i<4; i++) dbgprintf(" %02x", bind_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][i]);
        dbgprintf("\n");
        
        packet[0x00] = 0x18;
        packet[0x01] = bind_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][0];
        packet[0x02] = bind_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][1];
        packet[0x03] = bind_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][2];
        packet[0x04] = bind_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][3];
        packet[0x05] = 0x3c;
        packet[0x06] = 0x14;
        packet[0x07] = 0x01;
        packet[0x08] = 0xfd;
        packet[0x09] = 0x00;
        packet[0x0a] = bind_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][3];
        packet[0x0b] = 0x00;
        packet[0x0c] = 0x00;
        packet[0x0d] = 0x00;
        packet[0x0e] = 0x00;
        packet[0x0f] = 0x00;
        packet[0x10] = 0x00;
        packet[0x11] = 0x00;
        packet[0x12] = 0x00;
        packet[0x13] = 0x00;
        packet[0x14] = 0x00;
        packet[0x15] = 0x00;
        packet[0x16] = 0x00;
        packet[0x17] = 0x00;
        packet[0x18] = 0x00;
    } else {
        NRF24L01_WriteReg(NRF24L01_05_RF_CH, data_channels[Model.proto_opts[PROTOOPTS_PAYLOAD]][data_channel] + 2);
        dbgprintf("Sending DATA packet on channel 0x%02x\n", data_channels[Model.proto_opts[PROTOOPTS_PAYLOAD]][data_channel]);
        read_controls(&throttle, &rudder, &elevator, &aileron, &flags, &rudder_trim, &elevator_trim, &aileron_trim);
        // data: 1a e1 84 37 00 01 02 88 00 00 7f 7f 32 20 40 40 00 d0
        packet[0x01] = data_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][0];
        packet[0x02] = data_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][1];
        packet[0x03] = data_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][2];
        packet[0x04] = data_payloads[Model.proto_opts[PROTOOPTS_PAYLOAD]][3];
        packet[0x05] = 0x00;
        packet[0x06] = 0x01;
        packet[0x07] = 0x02;
        packet[0x08] = 0x88;
        packet[0x09] = 0x00;
        packet[0x0a] = 0x00;
        packet[0x0b] = 0x7f;
        packet[0x0c] = 0x7f;
        packet[0x0d] = 0x32;
        packet[0x0e] = 0x20;
        packet[0x0f] = 0x40;
        packet[0x10] = 0x40;
        packet[0x11] = 0x00;
        packet[0x12] = 0xd0;
        packet[0x13] = 0x00;
        packet[0x14] = 0x00;
        packet[0x15] = 0x00;
        packet[0x16] = 0x00;
        packet[0x17] = 0x00;
        packet[0x18] = 0x00;

        data_channel++;
        if (data_channel > 3) data_channel = 0;
/*
        packet[0] = throttle;
        packet[1] = rudder;
        packet[3] = elevator;
        packet[4] = aileron;
        packet[2] = elevator_trim;
        packet[5] = aileron_trim;
        packet[6] = rudder_trim;
        packet[7] = flags;
*/
    }

    crc_kermit = 0;
    for (int i=0; i <= 0x18; i++) {
      crc_kermit = update_crc_kermit(crc_kermit, (packet[i]));
    }

    dbgprintf("crc_kermit: 0x%04x\n", crc_kermit);
    crc_tabkermit_init = FALSE;

    packet[0x19] = (u8)(crc_kermit & 0x00ff);
    packet[0x1a] = (u8)((crc_kermit & 0xff00) >> 8);

    //packet[0x19] = (0xf3); // or 0xcf
    //packet[0x1a] = (0x3e); // or 0x7c

    // then shift the whole payload 12 bits to the right and insert the last address byte and the 4 bit trailer

    /* I have to draw this up to understand it:

    Byte 0x00: 0xAA = 1010 1010 -------- Preamble  ---
    Byte 0x01: 0xAA = 1010 1010 -----                |--- Preamble H1
    Byte 0x02: 0xAA = 1010 1010     |-----------------
    Byte 0x03: 0xE0 = 1110 0000     |--- Address   ------------
    Byte 0x04: 0xBF = 1011 1111     |                         |
    Byte 0x05: 0xA0 = 1010 0000------                         |--- Address H1
    Byte 0x06: 0x20 = 0010 0000 -------- First payload byte ---
    Byte 0x07: 0xA1 = 1010 0001 -------- 4 trailer bits and first nibble of packet array
    Byte 0x08: 0x8A = 1000 1010 -------- second and third nibble of packet array
    Byte 0x09: 0x9D = 1001 1101 ...
    ...
    Byte 0x17: 0x00 = 0000 0000
    Byte 0x13: 0x0? = 0000 ???? -------- last packet nibble and first nibble of CRC
    Byte 0x13: 0x?? = ???? ???? -------- second and third nibble of CRC
    Byte 0x14: 0x?x = ???? xxxx -------- last CRC nibble and nothing (should get ignored by the receiver)
*/


    // reverse bytes in packet
    for (u8 i=0; i < PACKET_SIZE; i++) {
      packet[i] = reverse(packet[i]);
    }

    dbgprintf("Packet before: ");
    for (u8 i=0; i < PACKET_SIZE; i++) {
      dbgprintf("%02x ", packet[i]);
    }

    dbgprintf("\nshifting:      ");
    u8 *shift = &packet[PACKET_SIZE-1];
    while (shift > (packet)) {
        //printf("\nshift: %02x (%02x) (%02x)", *shift, (*(shift-2)&0x0F)<<4, (*(shift-1)&0xF0)>>4);
        *shift = (*(shift-2)&0x0F)<<4 | (*(shift-1)&0xF0)>>4;
        //printf(" shifted: %02x", *shift);
        shift--;
    }
    packet[1] = 0xA0 | (*(packet+1)&0x0F); // trailer and first nibble of payload
    packet[0] = 0x20; // last address byte of H1 address


    dbgprintf("\nPacket after:  "); 
    for (u8 i=0; i <= PACKET_SIZE; i++) {
      dbgprintf("%02x ", packet[i]);
    }
    dbgprintf("\n\n");

    // clear packet status bits and TX FIFO
    NRF24L01_WriteReg(NRF24L01_07_STATUS, (BV(NRF24L01_07_TX_DS) | BV(NRF24L01_07_MAX_RT)));
    NRF24L01_FlushTx();

    NRF24L01_WritePayload(packet, PACKET_SIZE);

    //if (bind) NRF24L01_SetTxRxMode(RX_EN);

    ++packet_counter;

    // Check and adjust transmission power. We do this after
    // transmission to not bother with timeout after power
    // settings change -  we have plenty of time until next
    // packet.
    if (tx_power != Model.tx_power) {
        //Keep transmit power updated
        tx_power = Model.tx_power;
        NRF24L01_SetPower(tx_power);
    }
}


static void h1_init()
{
    dbgprintf("Initializing nRF24L01 on channel 0x%02x (%04dMHz)\n", bind_channels[Model.proto_opts[PROTOOPTS_PAYLOAD]], (bind_channels[Model.proto_opts[PROTOOPTS_PAYLOAD]] + 2402));
    NRF24L01_Initialize();

    // CRC, radio on
    NRF24L01_WriteReg(NRF24L01_00_CONFIG, BV(NRF24L01_00_PWR_UP)); // No CRC
    NRF24L01_WriteReg(NRF24L01_01_EN_AA, 0x00);      // EN_AA must be 0x00 to disable enhanced shockburst
    NRF24L01_WriteReg(NRF24L01_02_EN_RXADDR, 0x3F);  // Enable all data pipes
    NRF24L01_WriteReg(NRF24L01_03_SETUP_AW, 0x03);   // 5-byte RX/TX address
    NRF24L01_WriteReg(NRF24L01_04_SETUP_RETR, 0x00); // 500uS retransmit t/o, 10 tries
    NRF24L01_WriteReg(NRF24L01_05_RF_CH, bind_channels[Model.proto_opts[PROTOOPTS_PAYLOAD]] + 2);
    NRF24L01_SetBitrate(NRF24L01_BR_1M);             // 1Mbps
    NRF24L01_SetPower(Model.tx_power);
    NRF24L01_WriteReg(NRF24L01_07_STATUS, 0x70);     // Clear data ready, data sent, and retransmit
    // H1 Syncword = 0xe0bfa020, this is the last two preamble bytes and the first three bytes of the syncword (the last syncword byte is the first payload byte)
    u8 rx_tx_addr[] = {0xa0, 0xbf, 0xe0, 0xaa, 0xaa};
    NRF24L01_WriteRegisterMulti(NRF24L01_0A_RX_ADDR_P0, rx_tx_addr, 5);
    NRF24L01_WriteRegisterMulti(NRF24L01_10_TX_ADDR, rx_tx_addr, 5);

    // set up receiver
    NRF24L01_WriteReg(NRF24L01_11_RX_PW_P0, PACKET_SIZE);
}


MODULE_CALLTYPE
static u16 h1_callback()
{
    dbgprintf("h1_callback phase %d\n", phase);
    switch (phase) {
    case H1_INIT1:
        phase = H1_BIND2;
        //MUSIC_Play(MUSIC_TELEMALARM1);
        break;

    case H1_BIND2: // 
        if (counter == 0) {
            NRF24L01_SetTxRxMode(TX_EN);
            send_packet(1);
            NRF24L01_SetTxRxMode(RX_EN);
            counter = BIND_COUNT;
        } else {
            counter -= 1;
        }
        // no break!
    case H1_BIND3:
        switch (packet_ack()) {
        case PKT_PENDING:
            phase = H1_BIND2;
            return PACKET_CHKTIME;                 // no ack yet
        case PKT_ACKED:
            phase = H1_DATA;
            PROTOCOL_SetBindState(0);
            MUSIC_Play(MUSIC_DONE_BINDING);
            break;
        }
        break;

    case H1_DATA:
#ifdef H1_TELEMETRY
        //update_telemetry();
#endif
        send_packet(0);
        break;
    }
    return PACKET_PERIOD;                          // Packet every 5ms
}



static void initialize()
{
    CLOCK_StopTimer();
    tx_power = Model.tx_power;
    //initialize_rx_tx_addr();
    packet_counter = 0;
    flags = 0;

    h1_init();
    phase = H1_INIT1;


#ifdef H1_TELEMETRY
    memset(&Telemetry, 0, sizeof(Telemetry));
    TELEMETRY_SetType(TELEM_DSM);
#endif

    PROTOCOL_SetBindState(0xFFFFFFFF);
    CLOCK_StartTimer(INITIAL_WAIT, h1_callback);
}

const void *H1_Cmds(enum ProtoCmds cmd)
{
    switch(cmd) {
        case PROTOCMD_INIT:  initialize(); return 0;
        case PROTOCMD_DEINIT:
        case PROTOCMD_RESET:
            CLOCK_StopTimer();
            return (void *)(NRF24L01_Reset() ? 1L : -1L);
        case PROTOCMD_CHECK_AUTOBIND: return (void *)1L; // always Autobind
        case PROTOCMD_BIND:  initialize(); return 0;
        case PROTOCMD_NUMCHAN: return (void *) 6L; // A, E, T, R, enable flip, enable light
        case PROTOCMD_DEFAULT_NUMCHAN: return (void *)5L;
        case PROTOCMD_CURRENT_ID: return Model.fixed_id ? (void *)((unsigned long)Model.fixed_id) : 0;
        case PROTOCMD_GETOPTIONS: return Model.protocol == PROTOCOL_H1 ? h1_opts : 0;
#ifdef H1_TELEMETRY
        case PROTOCMD_TELEMETRYSTATE:
            return (void *)(long)(Model.proto_opts[PROTOOPTS_TELEMETRY] == TELEM_ON ? PROTO_TELEM_ON : PROTO_TELEM_OFF);
#else
        case PROTOCMD_TELEMETRYSTATE: return (void *)(long)PROTO_TELEM_UNSUPPORTED;
#endif
        case PROTOCMD_SET_TXPOWER:
            tx_power = Model.tx_power;
            NRF24L01_SetPower(tx_power);

        default: break;
    }
    return 0;
}
#endif

